/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2001
 *
 * Copyright (c) 2001 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "configuration.h"
#include "crl.h"

/* External imported variables */
extern OCSPD_CONFIG * ocspd_conf;

/* Functions */

void lookup_fail(char *name, char *tag) {
        syslog(LOG_ERR,"variable lookup failed for %s::%s\n",name,tag);
	return;
}

CONF *load_config( char *configfile, char **section ) {

	CONF *conf;
	long errorline= -1;

        if (configfile == NULL) configfile = getenv("OCSP_CONF");
        if (configfile == NULL) configfile = CONFIG_FILE;
        if (configfile == NULL) return(NULL);

	if( ocspd_conf->verbose )
        	syslog(LOG_INFO,"Using configuration from %s\n",configfile);

	conf = NCONF_new(NULL);
        if ( NCONF_load(conf, configfile, &errorline) <= 0 )
                {
                if (errorline <= 0)
                        syslog(LOG_ERR,
				"error loading the config file '%s'\n",
                                configfile);
                else
                        syslog(LOG_ERR,
				"error on line %ld of config file '%s'\n"
                                ,errorline, configfile);
                return(NULL);
                }

        if (*section == NULL) {
#ifdef OCSPD_ARCH_SOLARIS
		char sol_buf[1024];
#endif
                *section=NCONF_get_string(conf,BASE_SECTION,ENV_DEFAULT_OCSPD);
                if (*section == NULL) {
                        lookup_fail(BASE_SECTION,ENV_DEFAULT_OCSPD);
                        return(NULL);
                }
#ifndef OCSPD_ARCH_SOLARIS
		setenv( ENV_SECTION, *section, 1);
#else
		sprintf(sol_buf, "ENV_SECTION=%s", (char *) section);
		putenv( sol_buf );
#endif
		if( ocspd_conf->verbose )
			syslog(LOG_INFO,"section set to %s", *section );
        }

	return(conf);
}

int ocspd_load_ca_section ( OCSPD_CONFIG *conf, char *dbms_section ) {
        STACK_OF(CONF_VALUE) *nval = NULL;
        CONF_VALUE *val = NULL;

	char *crlUrl_s = NULL;
	char *compromised_s = NULL;
	/*
	char *crl_dn_s = NULL;
	X509_CRL *crl_crl = NULL;
	*/

	CRL_DATA crl_data;

	/*
	char *caUrl_key = NULL;
	*/
	char *caUrl_s = NULL;
	char *servUrl_s = NULL;
	/*
	char *tmp_s = NULL;
	*/

	char *ca_section = NULL;
	CA_LIST_ENTRY *ca = NULL;
	/*
	CA_LIST_ENTRY tmp_ca;
	*/

        int i, ret;
	/*
	size_t list_size = 0;
	char buf[2048];
	*/
	URL *serv_cert_url = NULL;

	/* Check for the existance of the dbms_section in config file */
        if (!(nval = NCONF_get_section(conf->conf, dbms_section)))
		return 0;

	if( ocspd_conf->verbose )
		syslog(LOG_INFO, "Number of CAs in configuration is %d",
			sk_CONF_VALUE_num(nval));

	/* At maximum we will have the number of CAs equal to the
	   entries in the dbms_section (each correasponds to a CA */
	conf->ca_list = (CA_LIST_ENTRY **) OPENSSL_malloc(
		sizeof(CA_LIST_ENTRY *) * sk_CONF_VALUE_num(nval));

	/* Check for memory allocation to be succesfull */
	if( conf->ca_list == NULL ) {
		syslog( LOG_ERR, "%s:%d Memory allocation error.",
			__FILE__, __LINE__ );
		return 0;
	}
	for( i = 0; i < sk_CONF_VALUE_num(nval); i++ ) {
		conf->ca_list[i] = (CA_LIST_ENTRY *)
			OPENSSL_malloc( sizeof( CA_LIST_ENTRY ));
	}

	/* Initialise the CA counter */
	conf->ca_list_len = 0;

	/* Now we go through the whole dbms_section and we try to
	   get all the crl data for every listed CA */
        for (i = 0; i < sk_CONF_VALUE_num(nval); i++) {
#ifdef OCSPD_ARCH_SOLARIS 
		char sol_buf[1024];
#endif
                val = sk_CONF_VALUE_value(nval, i);

		if( strstr( val->name, ENV_OCSPD_CA_SECTION ) != NULL ) {
			if( (ca_section = strchr(val->value, '@')) == NULL ) {
				syslog(LOG_ERR, 
					"Error, ca section value should begin with @ (%s=%s)", 
					val->name, val->value);
					continue;
			}
			/* The first character should be a '@' so we need
			   to inc the ca_section pointer */
			ca_section++;

#ifndef OCSPD_ARCH_SOLARIS 
			setenv( ENV_SECTION, ca_section, 1);
#else
			sprintf(sol_buf, "ENV_SECTION=%s", ca_section);
			putenv( sol_buf );
#endif

			/* Get the char * with the value for the CA URL */
		        if ((caUrl_s = NCONF_get_string(conf->conf, 
				ca_section, ENV_OCSPD_CACERT_URL)) == NULL) {

               			if( ocspd_conf->verbose )
					lookup_fail(ca_section, 
							ENV_OCSPD_CACERT_URL);

				continue;
		      	}; 


			/* Let's point to the CA_LIST entry */
			ca = conf->ca_list[conf->ca_list_len];

			/* Get the CA parsed url */
			ca->url = getParsedUrl( caUrl_s );

			/* Error, can not parse url data */
			if( ! ca->url ) {
				syslog( LOG_ERR, "Can not parse CA url in section %s (%s)",
					ca_section, caUrl_s);
				return 0;
			}
			/* This is used if LDAP is used for CA certificate
			   retrieval */
        		ca->url->dn = NCONF_get_string(conf->conf, 
					ca_section, ENV_OCSPD_CACERT_ENTRY_DN);
        		ca->url->attr = NCONF_get_string(conf->conf, 
				ca_section, ENV_OCSPD_CACERT_ENTRY_ATTR);

			/* Get the CA entry, usually the Base DN of the LDAP */
			if( ca->url->proto == OCSP_CRL_PROTO_LDAP ){
                                if( ocspd_conf->verbose )
                                        syslog(LOG_INFO,
                                                "Using LDAP protocol for CA retrivial");

                                if( !(ca->url->attr) ||
                                                !(ca->url->dn) ) {
                                        syslog( LOG_ERR,
                                                "%s and %s are required while using LDAP!",
                                                ENV_OCSPD_CACERT_ENTRY_DN,
                                                ENV_OCSPD_CACERT_ENTRY_ATTR );

                                        return 0;
                                }
                        }

			/* Get the char * with the value for the OCSP URL */
		        if ((servUrl_s = NCONF_get_string(conf->conf, 
				ca_section, ENV_OCSPD_SERVER_CERT_URL))==NULL) {

               			if( ocspd_conf->verbose )
					lookup_fail(ca_section, 
						ENV_OCSPD_SERVER_CERT_URL);

				syslog(LOG_INFO, "WARNING: No server cert "
					"found for %s, skipping CA "
					"configuration!", ca_section );
				continue;
		      	}; 

			/* Get the OCSP certificate issued by THIS CA */
			if( servUrl_s ) {
				serv_cert_url = getParsedUrl( servUrl_s );
				if( serv_cert_url != NULL ) {
					ca->ocsp_cert = ocspd_get_cert (serv_cert_url);
					URL_free( serv_cert_url );
				}
			}

			/* Get the CA certificate */
			ca->cert = ocspd_get_ca(ca->url);
			if(!ca->cert || (sk_X509_num(ca->cert) < 1)) {
				if(ca->cert) sk_X509_free(ca->cert);
				syslog(LOG_ERR, "Error loading CA URL data "
						"for %s, skipping CA config",
						ca_section );
				continue;
			} else {
				if(ocspd_conf->verbose)
				   syslog( LOG_INFO, 
				      "CA CERT for %s loaded successfully.",
				      ca_section );
			}

			if((ca->cid = ocspd_CA_ENTRY_CERTID_new ( ca->cert, 
						conf->digest)) == NULL ) {
				syslog( LOG_ERR,
					"%s:%d CA List structure init error.",
						__FILE__, __LINE__ );
				continue;
			}

			if(ocspd_conf->verbose)
				syslog(LOG_INFO, 
					"CA %s List Entry added (CA list "
						"num %ld)", ca_section,
							conf->ca_list_len);

			/* First CA data correctly loaded */
			conf->ca_list_len++;

			/* Check if the CA has been marked as compromised */
		        if ((compromised_s = NCONF_get_string(conf->conf, 
				ca_section, ENV_OCSPD_CA_COMPROMISED))==NULL) {
					ca->compromised = 0;
			} else {
				if( strncmp_nocase( compromised_s, 
							"y", 1 ) == 0 ) {
					ca->compromised = 1;
				} else {
					ca->compromised = 0;
				}
			}

			/* Now process the CRL data */
		        if ((crlUrl_s = NCONF_get_string(conf->conf, 
					ca_section, ENV_OCSPD_CRL_URL))==NULL) {

				X509 *x = NULL;

				x = sk_X509_value(ca->cert, 0);

				if( ocspd_conf->verbose )
					lookup_fail(ca_section,
							ENV_OCSPD_CRL_URL);

				/* If the CDP extension is present in
				   the CA CERTIFICATE and the cert is
				   self-signed */
				if( X509_NAME_cmp(
					X509_get_subject_name(x), 
					X509_get_issuer_name(x)) == 0 ) {

				   if( ocspd_conf->verbose )
				      syslog(LOG_INFO, 
						"Root CA found, check for "
						"CDP extension");

				   if( (crlUrl_s=ocspd_parse_cdp_ext(ca->cert)) != NULL) {
					if (ocspd_conf->verbose)
						syslog( LOG_INFO, "Using CDP "
						   "extension from CA cert for"
							" %s", ca_section );
				   } else {
					syslog( LOG_ERR, "ERROR: no CRL url"
						"available (CONFIG, CA cert "
						"CDP extension) for %s",
						ca_section );
				   }
				}
		      	}; 

			/* Set the ID of the CA */
			ca->ca_id = ca_section;

			memset(&crl_data, '\x0', sizeof(crl_data));

			/* Parse the CRL url */
			crl_data.url = getParsedUrl( crlUrl_s );

			/* Error, can not get crl data */
			if( ! crl_data.url ) {
				syslog( LOG_ERR, 
					"Can not parse CRL url @%s (%s)",
					ca_section, crlUrl_s);
				return 0;
			}

			ca->crl_url = crl_data.url;

        		crl_data.url->dn = NCONF_get_string(conf->conf, 
				ca_section, ENV_OCSPD_CRL_ENTRY_DN);
        		crl_data.url->attr = NCONF_get_string(conf->conf, 
				ca_section, ENV_OCSPD_CRL_ENTRY_ATTR);

			/* Get the CRL entry, usually the Base DN of the LDAP */
			if( crl_data.url->proto == OCSP_CRL_PROTO_LDAP ){
				if( ocspd_conf->verbose )
					syslog(LOG_INFO,
						"Using LDAP protocol for CRL "
						"retrivial for %s", ca_section);

				if( !(crl_data.url->attr) || 
						!(crl_data.url->dn) ) {
					syslog( LOG_ERR,
						"%s and %s are required while using LDAP!",
						ENV_OCSPD_CRL_ENTRY_DN,
						ENV_OCSPD_CRL_ENTRY_ATTR );
				
					return 0;
				}
			}

			if((ca->crl = ocspd_get_crl( crl_data.url )) == NULL) {
				syslog(LOG_ERR, "ERROR::Can not load CRL for "
					"[ %s ]", ca->ca_id );
			} else {
				if( ocspd_conf->verbose ) {
					syslog(LOG_INFO, "CRL loaded [ %s ]",
						ca_section );
				}

				/* Let's check the CRL against the CA certificate */
				if( (ret = check_crl( ca->crl, ca->cert )) < 1 ) {
					syslog( LOG_ERR, "CRL/CA check error [ %s:%d ]",
						ca_section, ret );
				}

				/* Now we copy the lastUpdate and nextUpdate fields */
				if (ca->crl) {
					const ASN1_TIME *lastup;
					const ASN1_TIME *nextup;

#if OPENSSL_VERSION_NUMBER < 0x10100005L
					lastup = X509_CRL_get_lastUpdate(ca->crl);
					nextup = X509_CRL_get_nextUpdate(ca->crl);
#else
					lastup = X509_CRL_get0_lastUpdate(ca->crl);
					nextup = X509_CRL_get0_nextUpdate(ca->crl);
#endif

					ca->lastUpdate =
					    (ASN1_TIME *) ASN1_STRING_dup(
					    (ASN1_STRING *) lastup);
					ca->nextUpdate =
					    (ASN1_TIME *) ASN1_STRING_dup(
					    (ASN1_STRING *) nextup);
				}

				ca->crl_status = check_crl_validity ( ca );

				if((ocspd_conf->verbose) && 
						(ca->crl_status == CRL_OK))
					syslog(LOG_INFO, "INFO::CRL for %s "
						"is Valid", ca_section );

				/* Let's get the CRLs entries, if any */
				if( ocspd_build_crl_entries_list ( ca, 
							ca->crl ) == NULL ) { 
					syslog(LOG_ERR, "No Entries for CRL (@%s)",
						ca_section );
				};

				syslog( LOG_INFO, "CRL loaded for %s", 
					ca_section );
			}
		}

        }

	return 1;
}

int ocspd_reload_all_ca ( OCSPD_CONFIG *conf ) {

	/*
	int cnt, ret;
	*/
	int i=0;
	CA_LIST_ENTRY *ca = NULL;

	for( i = 0; i < conf->ca_list_len; i++ ) {

		ca = conf->ca_list[i];

		/* Let's free the CA certs list, if present */
		if( ca->cert ) {
			sk_X509_pop_free(ca->cert, X509_free );
		}
		/* Get the CA certificate */
		ca->cert = ocspd_get_ca(ca->url);
		if(!ca->cert || !sk_X509_num(ca->cert)) {
			syslog(LOG_ERR, "Error loading CA URL data.");
			continue;
		} else {
			if(ocspd_conf->verbose)
				syslog( LOG_INFO,
					"CA CERT for %s loaded successfully.",
					ca->ca_id );
		}

		if((ca->cid = ocspd_CA_ENTRY_CERTID_new ( ca->cert,
						conf->digest)) == NULL ) {
			syslog( LOG_ERR, "CA List structure init error.");
			continue;
		}

	}

	return 1;
}

